"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const ObjectType_1 = require("../definition/ObjectType");
const global_1 = require("../global");
const Field_1 = require("../reference/Field");
const ast_1 = require("../sdl/ast");
const coerceType_1 = require("../type-expression/coerceType");
const structure_1 = require("../type-expression/structure");
function defineConnection(options) {
    return function registerConnectionMetadata(definitionClass) {
        const storage = global_1.getGlobalMetadataStorage();
        storage.deferRegister(() => {
            const connectionName = options.typeName || definitionClass.name;
            if (!connectionName) {
                throw new Error(`Should provide connection typeName`);
            }
            let edge;
            const node = coerceType_1.coerceType(options.node);
            const nodeTypeName = node.getTypeName({ storage, kind: 'output' });
            if (options.edge) {
                edge = coerceType_1.coerceType(options.edge);
            }
            else {
                const edgeTypeName = `${nodeTypeName}Edge`;
                defineEdge({ typeName: edgeTypeName, node })(null);
                edge = coerceType_1.coerceType(edgeTypeName);
            }
            storage.registerMetadata([
                new ObjectType_1.ObjectType({
                    definitionClass,
                    definitionName: connectionName,
                    description: 'A connection to a list of items.'
                }),
                new Field_1.Field({
                    source: definitionClass,
                    target: structure_1.NonNull.of(coerceType_1.coerceType(pageInfoType)),
                    fieldName: 'pageInfo',
                    description: 'Information to aid in pagination.',
                    args: [],
                    resolver: ast_1.bindStaticResolver(definitionClass, 'pageInfo') || graphql_1.defaultFieldResolver,
                }),
                new Field_1.Field({
                    source: definitionClass,
                    target: structure_1.NonNull.of(structure_1.List.of(structure_1.NonNull.of(edge))),
                    fieldName: 'edges',
                    description: 'A list of edges.',
                    args: [],
                    resolver: ast_1.bindStaticResolver(definitionClass, 'edges') || graphql_1.defaultFieldResolver,
                }),
            ]);
        });
    };
}
exports.defineConnection = defineConnection;
function defineEdge(options) {
    return function registerEdgeMetadata(definitionClass) {
        const storage = global_1.getGlobalMetadataStorage();
        storage.deferRegister(() => {
            const node = coerceType_1.coerceType(options.node);
            const nodeTypeName = node.getTypeName({ storage, kind: 'output' });
            const edgeName = options.typeName
                || (definitionClass && definitionClass.name)
                || `${nodeTypeName}Edge`;
            storage.registerMetadata([
                new ObjectType_1.ObjectType({
                    definitionClass,
                    definitionName: edgeName,
                    description: 'An edge in a connection.',
                }),
                new Field_1.Field({
                    source: definitionClass || edgeName,
                    target: structure_1.NonNull.of(node),
                    fieldName: 'node',
                    description: 'The item at the end of the edge',
                    args: [],
                    resolver: definitionClass && ast_1.bindStaticResolver(definitionClass, 'pageInfo') || graphql_1.defaultFieldResolver,
                }),
                new Field_1.Field({
                    source: definitionClass || edgeName,
                    target: structure_1.NonNull.of(coerceType_1.coerceType('String')),
                    fieldName: 'cursor',
                    description: 'A cursor for use in pagination',
                    args: [],
                    resolver: definitionClass && ast_1.bindStaticResolver(definitionClass, 'cursor') || graphql_1.defaultFieldResolver,
                }),
            ]);
        });
    };
}
exports.defineEdge = defineEdge;
const pageInfoType = new graphql_1.GraphQLObjectType({
    name: 'PageInfo',
    description: 'Information about pagination in a connection.',
    fields: () => ({
        hasNextPage: {
            type: new graphql_1.GraphQLNonNull(graphql_1.GraphQLBoolean),
            description: 'When paginating forwards, are there more items?'
        },
        hasPreviousPage: {
            type: new graphql_1.GraphQLNonNull(graphql_1.GraphQLBoolean),
            description: 'When paginating backwards, are there more items?'
        },
        startCursor: {
            type: graphql_1.GraphQLString,
            description: 'When paginating backwards, the cursor to continue.'
        },
        endCursor: {
            type: graphql_1.GraphQLString,
            description: 'When paginating forwards, the cursor to continue.'
        },
    })
});
//# sourceMappingURL=connection.js.map