"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const SubscriptionType_1 = require("../definition/SubscriptionType");
const ObjectType_1 = require("../definition/ObjectType");
const InterfaceType_1 = require("../definition/InterfaceType");
const InputType_1 = require("../definition/InputType");
const coerceType_1 = require("../type-expression/coerceType");
const structure_1 = require("../type-expression/structure");
const Field_1 = require("../reference/Field");
const InputField_1 = require("../reference/InputField");
const Implement_1 = require("../reference/Implement");
class DefinitionParser {
    constructor(rootNode, subsMap) {
        this.rootNode = rootNode;
        this.subsMap = subsMap;
    }
    parseWithResolverMap(definitionClass) {
        const { rootNode } = this;
        this.metadata = [];
        if (rootNode.kind === 'ObjectTypeDefinition') {
            this.handleObjectTypeDefinition(rootNode, definitionClass);
        }
        else if (rootNode.kind === 'ObjectTypeExtension') {
            this.handleObjectTypeExtension(rootNode, definitionClass);
        }
        else if (rootNode.kind === 'InterfaceTypeDefinition') {
            this.handleInterfaceTypeDefinition(rootNode, definitionClass);
        }
        else if (rootNode.kind === 'InterfaceTypeExtension') {
            this.handleInterfaceTypeExtension(rootNode, definitionClass);
        }
        else if (rootNode.kind === 'InputObjectTypeDefinition') {
            this.handleInputObjectTypeDefinition(rootNode, definitionClass);
        }
        else if (rootNode.kind === 'InputObjectTypeExtension') {
            this.handleInputObjectExtension(rootNode, definitionClass);
        }
        else {
            throw new Error(`Node type not supported: ${rootNode.kind}`);
        }
        return this.metadata;
    }
    completeTypeExpression(typeNode) {
        const { subsMap } = this;
        if (typeNode.kind === 'ListType') {
            return new structure_1.List(this.completeTypeExpression(typeNode.type));
        }
        else if (typeNode.kind === 'NonNullType') {
            return new structure_1.NonNull(this.completeTypeExpression(typeNode.type));
        }
        const subType = subsMap[typeNode.name.value];
        return coerceType_1.coerceType(subType || typeNode.name.value);
    }
    handleObjectTypeDefinition(rootNode, definitionClass) {
        const { name, interfaces, fields, description } = rootNode;
        if (fields) {
            fields.forEach(fieldNode => this.appendFieldMetadataConfig(fieldNode, definitionClass));
        }
        if (interfaces) {
            interfaces.forEach(interfaceNode => this.appendImplementTypeExpression(interfaceNode, definitionClass));
        }
        let metadata;
        if (name.value === 'Subscription') {
            metadata = new SubscriptionType_1.SubscriptionType({
                definitionClass,
                definitionName: name.value,
                description: description && description.value,
                directives: rootNode.directives && this.completeDirectives(rootNode.directives),
            });
        }
        else {
            metadata = new ObjectType_1.ObjectType({
                definitionClass,
                definitionName: name.value,
                description: description && description.value,
                directives: rootNode.directives && this.completeDirectives(rootNode.directives),
            });
        }
        this.metadata.push(metadata);
    }
    handleObjectTypeExtension(rootNode, definitionClass) {
        const { name, interfaces, fields } = rootNode;
        if (fields) {
            fields.forEach(fieldNode => this.appendFieldMetadataConfig(fieldNode, definitionClass, name.value));
        }
        if (interfaces) {
            interfaces.forEach(interfaceNode => this.appendImplementTypeExpression(interfaceNode, definitionClass, name.value));
        }
    }
    handleInterfaceTypeDefinition(rootNode, definitionClass) {
        const { name, description, fields } = rootNode;
        if (fields) {
            fields.forEach(fieldNode => this.appendFieldMetadataConfig(fieldNode, definitionClass));
        }
        this.metadata.push(new InterfaceType_1.InterfaceType({
            definitionClass,
            definitionName: name.value,
            description: description && description.value,
            directives: rootNode.directives && this.completeDirectives(rootNode.directives),
        }));
    }
    handleInterfaceTypeExtension(rootNode, definitionClass) {
        const { name, fields } = rootNode;
        if (fields) {
            fields.forEach(fieldNode => this.appendFieldMetadataConfig(fieldNode, definitionClass, name.value));
        }
    }
    handleInputObjectTypeDefinition(node, definitionClass) {
        const { name, description, fields } = node;
        if (fields) {
            fields.forEach(fieldNode => this.appendInputFieldMetadataConfig(fieldNode, definitionClass));
        }
        this.metadata.push(new InputType_1.InputType({
            definitionClass,
            definitionName: name.value,
            description: description && description.value,
            directives: node.directives && this.completeDirectives(node.directives),
        }));
    }
    handleInputObjectExtension(node, definitionClass) {
        const { name, fields } = node;
        if (fields) {
            fields.forEach(fieldNode => this.appendInputFieldMetadataConfig(fieldNode, definitionClass, name.value));
        }
    }
    appendFieldMetadataConfig(node, definitionClass, extendingTypeName) {
        const { name, type, description, directives, arguments: args } = node;
        const argumentRefs = args && args.map(argumentNode => (this.createInputField(argumentNode, definitionClass)));
        const resolver = bindStaticResolver(definitionClass, name.value) || graphql_1.defaultFieldResolver;
        this.metadata.push(new Field_1.Field({
            source: extendingTypeName || definitionClass,
            target: this.completeTypeExpression(type),
            fieldName: name.value,
            args: argumentRefs || [],
            description: description && description.value,
            directives: directives && this.completeDirectives(directives),
            extendingTypeName,
            resolver,
        }));
    }
    createInputField(node, definitionClass, extendingTypeName) {
        const { name, type, description, defaultValue, directives } = node;
        return new InputField_1.InputField({
            source: extendingTypeName || definitionClass,
            target: this.completeTypeExpression(type),
            fieldName: name.value,
            directives: directives && this.completeDirectives(directives),
            defaultValue: defaultValue && this.completeValueNode(defaultValue),
            description: description && description.value,
            extendingTypeName,
        });
    }
    appendInputFieldMetadataConfig(node, definitionClass, extendingTypeName) {
        const field = this.createInputField(node, definitionClass, extendingTypeName);
        this.metadata.push(field);
    }
    appendImplementTypeExpression(node, definitionClass, extendingTypeName) {
        this.metadata.push(new Implement_1.Implement({
            source: extendingTypeName || definitionClass,
            target: this.completeTypeExpression(node),
        }));
    }
    completeDirectives(directiveNodes) {
        return directiveNodes.reduce((results, node) => {
            if (node.arguments instanceof Array) {
                results[node.name.value] = this.completeArgumentsOrObjectFields(node.arguments);
            }
            else {
                results[node.name.value] = {};
            }
            return results;
        }, {});
    }
    completeArgumentsOrObjectFields(nodes) {
        return nodes.reduce((results, node) => {
            results[node.name.value] = this.completeValueNode(node.value);
            return results;
        }, {});
    }
    completeValueNode(node) {
        if (node.kind === 'ObjectValue') {
            return this.completeArgumentsOrObjectFields(node.fields);
        }
        else if (node.kind === 'NullValue') {
            return null;
        }
        else if (node.kind === 'ListValue') {
            return node.values.map(this.completeValueNode.bind(this));
        }
        else if (node.kind === 'Variable') {
            throw new Error(`Cannot use variable in schema directives: ${node.name}`);
        }
        else if (node.kind === 'IntValue' || node.kind === 'FloatValue') {
            return Number(node.value);
        }
        return node.value;
    }
}
exports.DefinitionParser = DefinitionParser;
function bindStaticResolver(definitionClass, fieldName) {
    const maybeStaticResolver = definitionClass[fieldName];
    return maybeStaticResolver instanceof Function
        ? maybeStaticResolver.bind(definitionClass)
        : null;
}
exports.bindStaticResolver = bindStaticResolver;
//# sourceMappingURL=ast.js.map